@@ -22,6 +22,7 @@
 #include "cpu.h"
 #include "lazyflags.h"
 #include "callback.h"
+#include "paging.h"
 #include "pic.h"
 #include "fpu.h"
 
@@ -39,17 +40,6 @@ extern bool ignore_opcode_63;
 #include "debug.h"
 #endif
 
-#if (!C_CORE_INLINE)
-#define LoadMb(off) mem_readb(off)
-#define LoadMw(off) mem_readw(off)
-#define LoadMd(off) mem_readd(off)
-#define LoadMq(off) ((uint64_t)((uint64_t)mem_readd(off+4)<<32 | (uint64_t)mem_readd(off)))
-#define SaveMb(off,val)	mem_writeb(off,val)
-#define SaveMw(off,val)	mem_writew(off,val)
-#define SaveMd(off,val)	mem_writed(off,val)
-#define SaveMq(off,val) {mem_writed(off,val&0xffffffff);mem_writed(off+4,(val>>32)&0xffffffff);}
-#else 
-#include "paging.h"
 #define LoadMb(off) mem_readb_inline(off)
 #define LoadMw(off) mem_readw_inline(off)
 #define LoadMd(off) mem_readd_inline(off)
@@ -58,7 +48,6 @@ extern bool ignore_opcode_63;
 #define SaveMw(off,val)	mem_writew_inline(off,val)
 #define SaveMd(off,val)	mem_writed_inline(off,val)
 #define SaveMq(off,val) {mem_writed_inline(off,val&0xffffffff);mem_writed_inline(off+4,(val>>32)&0xffffffff);}
-#endif
 
 extern Bitu cycle_count;
 
@@ -191,15 +180,17 @@ void CPU_Core_Prefetch_reset(void) {
 Bits CPU_Core_Prefetch_Run(void) {
 	bool invalidate_pq=false;
 
-    if (CPU_Cycles <= 0)
-	    return CBRET_NONE;
+	if (CPU_Cycles <= 0)
+		return CBRET_NONE;
 
-    // FIXME: This makes 8086 4-byte prefetch queue impossible to emulate.
-    //        The best way to accomplish this is to have an alternate version
-    //        of this prefetch queue for 286 or lower that fetches in 16-bit
-    //        WORDs instead of 32-bit WORDs.
-    pq_limit = (max(CPU_PrefetchQueueSize,(unsigned int)(4ul + prefetch_unit)) + prefetch_unit - 1ul) & (~(prefetch_unit-1ul));
-    pq_reload = min(pq_limit,(Bitu)8u);
+	const Bitu init_cycles = CPU_Cycles;
+
+	// FIXME: This makes 8086 4-byte prefetch queue impossible to emulate.
+	//        The best way to accomplish this is to have an alternate version
+	//        of this prefetch queue for 286 or lower that fetches in 16-bit
+	//        WORDs instead of 32-bit WORDs.
+	pq_limit = (max(CPU_PrefetchQueueSize,(unsigned int)(4ul + prefetch_unit)) + prefetch_unit - 1ul) & (~(prefetch_unit-1ul));
+	pq_reload = min(pq_limit,(Bitu)8u);
 
 	while (CPU_Cycles-->0) {
 		if (invalidate_pq) {
@@ -294,6 +285,7 @@ restart_opcode:
     }
 #endif
 
+decode_stop_at_instruction:
 	FillFlags();
 	return CBRET_NONE;
 decode_end:
