@@ -30,6 +30,7 @@
 extern int maxfcb;
 extern bool gbk, chinasea;
 extern Bitu DOS_PRIVATE_SEGMENT_Size;
+extern uint16_t desired_ems_segment;
 #if defined(USE_TTF)
 extern bool ttf_dosv;
 #endif
@@ -73,6 +74,11 @@ void DOS_GetMemory_reset() {
 	dos_memseg = 0;
 }
 
+void DOS_FreeTableMemory()
+{
+    dos_memseg = DOS_PRIVATE_SEGMENT;
+}
+
 void DOS_GetMemory_reinit() {
     DOS_GetMemory_unmapped = false;
     DOS_GetMemory_reset();
@@ -94,38 +100,48 @@ bool DOS_User_Wants_UMBs() {
     return section->Get_bool("umb");
 }
 
+bool EMS_Active(void);
+void Update_Get_Desired_Segment(void);
+
 void DOS_GetMemory_Choose() {
 	if (DOS_PRIVATE_SEGMENT == 0) {
-        /* DOSBox-X non-compatible: Position ourself just past the VGA BIOS */
-        /* NTS: Code has been arranged so that DOS kernel init follows BIOS INT10h init */
-        DOS_PRIVATE_SEGMENT=(uint16_t)VGA_BIOS_SEG_END;
-        DOS_PRIVATE_SEGMENT_END= (uint16_t)(DOS_PRIVATE_SEGMENT + DOS_PRIVATE_SEGMENT_Size);
+		/* DOSBox-X non-compatible: Position ourself just past the VGA BIOS */
+		/* NTS: Code has been arranged so that DOS kernel init follows BIOS INT10h init */
+		DOS_PRIVATE_SEGMENT=(uint16_t)VGA_BIOS_SEG_END;
 
-        if (IS_PC98_ARCH) {
-            bool PC98_FM_SoundBios_Enabled(void);
+		if (desired_ems_segment == 0) Update_Get_Desired_Segment();
 
-            /* Do not let the private segment overlap with anything else after segment C800:0000 including the SOUND ROM at CC00:0000.
-             * Limiting to 32KB also leaves room for UMBs if enabled between C800:0000 and the EMS page frame at (usually) D000:0000 */
-            unsigned int limit = 0xD000;
+		if (DOS_PRIVATE_SEGMENT == desired_ems_segment) {
+			DOS_PRIVATE_SEGMENT += 0x1000; // FIXME
+		}
 
-            if (PC98_FM_SoundBios_Enabled()) {
-                // TODO: What about sound BIOSes larger than 16KB?
-                if (limit > 0xCC00)
-                    limit = 0xCC00;
-            }
+		DOS_PRIVATE_SEGMENT_END= (uint16_t)(DOS_PRIVATE_SEGMENT + DOS_PRIVATE_SEGMENT_Size);
 
-            if (DOS_User_Wants_UMBs()) {
-                // leave room for UMBs, things are cramped a bit in PC-98 mode
-                if (limit > 0xC600)
-                    limit = 0xC600;
-            }
+		if (IS_PC98_ARCH && (desired_ems_segment == 0xD000)) {
+			bool PC98_FM_SoundBios_Enabled(void);
 
-            if (DOS_PRIVATE_SEGMENT_END > limit)
-                DOS_PRIVATE_SEGMENT_END = limit;
+			/* Do not let the private segment overlap with anything else after segment C800:0000 including the SOUND ROM at CC00:0000.
+			 * Limiting to 32KB also leaves room for UMBs if enabled between C800:0000 and the EMS page frame at (usually) D000:0000 */
+			unsigned int limit = 0xD000;
 
-            if (DOS_PRIVATE_SEGMENT >= DOS_PRIVATE_SEGMENT_END)
-                E_Exit("Insufficient room in upper memory area for private area");
-        }
+			if (PC98_FM_SoundBios_Enabled()) {
+				// TODO: What about sound BIOSes larger than 16KB?
+				if (limit > 0xCC00)
+					limit = 0xCC00;
+			}
+
+			if (DOS_User_Wants_UMBs()) {
+				// leave room for UMBs, things are cramped a bit in PC-98 mode
+				if (limit > 0xC600)
+					limit = 0xC600;
+			}
+
+			if (DOS_PRIVATE_SEGMENT_END > limit)
+				DOS_PRIVATE_SEGMENT_END = limit;
+
+			if (DOS_PRIVATE_SEGMENT >= DOS_PRIVATE_SEGMENT_END)
+				E_Exit("Insufficient room in upper memory area for private area");
+		}
 
 		if (DOS_PRIVATE_SEGMENT >= 0xA000) {
 			memset(GetMemBase()+((Bitu)DOS_PRIVATE_SEGMENT<<4u),0x00,(Bitu)(DOS_PRIVATE_SEGMENT_END-DOS_PRIVATE_SEGMENT)<<4u);
@@ -216,8 +232,6 @@ PhysPt DOS_Get_DPB(unsigned int dos_drive) {
 }
 
 void SetupDBCSTable() {
-    if (dos.loaded_codepage==950&&!chinasea) makestdcp950table();
-    else if (dos.loaded_codepage==951&&chinasea) makeseacp951table();
     if (enable_dbcs_tables) {
         if (!dos.tables.dbcs) dos.tables.dbcs=RealMake(DOS_GetMemory(12,"dos.tables.dbcs"),0);
 
@@ -252,6 +266,8 @@ void SetupDBCSTable() {
     else {
         dos.tables.dbcs=0;
     }
+    if(dos.loaded_codepage == 950 && !chinasea) makestdcp950table();
+    else if(dos.loaded_codepage == 951 && chinasea) makeseacp951table();
 }
 
 uint16_t seg_win_startup_info;
@@ -266,12 +282,6 @@ void DOS_SetupTables(void) {
 	/* create SDA */
 	DOS_SDA(DOS_SDA_SEG,0).Init();
 
-	/* Some weird files >20 detection routine */
-	/* Possibly obsolete when SFT is properly handled */
-	real_writed(DOS_CONSTRING_SEG,0x0a,0x204e4f43);
-	real_writed(DOS_CONSTRING_SEG,0x1a,0x204e4f43);
-	real_writed(DOS_CONSTRING_SEG,0x2a,0x204e4f43);
-
 	/* create a CON device driver */
 	if(IS_DOSV) {
 		seg = DOS_GetMemory(2, "device $IBMADSP");
@@ -384,7 +394,12 @@ void DOS_SetupTables(void) {
 
 	/* Create a fake DPB */
 	dos.tables.dpb=DOS_GetMemory(((DOS_DRIVES*dos.tables.dpb_size)+15u)/16u,"dos.tables.dpb");
-    dos.tables.mediaid_offset=0x17;	//Media ID offset in DPB (MS-DOS 4.x-6.x)
+
+	if (dos.version.major >= 4)
+		dos.tables.mediaid_offset=0x17;	//Media ID offset in DPB (MS-DOS 4.x-6.x)
+	else
+		dos.tables.mediaid_offset=0x16;	//Media ID offset in DPB (MS-DOS 1.x-3.x)
+
 	dos.tables.mediaid=RealMake(dos.tables.dpb,dos.tables.mediaid_offset);
 	for (i=0;i<DOS_DRIVES;i++) {
         real_writeb(dos.tables.dpb,i*dos.tables.dpb_size,(uint8_t)i);             // drive number
@@ -392,15 +407,15 @@ void DOS_SetupTables(void) {
         real_writew(dos.tables.dpb,i*dos.tables.dpb_size+2,0x0200);     // bytes per sector
         real_writew(dos.tables.dpb,i*dos.tables.dpb_size+6,0x0001);     // reserved sectors at the beginning of the drive
         mem_writew(Real2Phys(dos.tables.mediaid)+i*dos.tables.dpb_size,0u);
-        real_writew(dos.tables.dpb,i*dos.tables.dpb_size+0x1F,0xFFFF);      // number of free clusters or 0xFFFF if unknown
+        real_writew(dos.tables.dpb,i*dos.tables.dpb_size+(dos.version.major>=4?0x1F:0x1E),0xFFFF);      // number of free clusters or 0xFFFF if unknown
 
         // next DPB pointer
         if ((i+1) < DOS_DRIVES)
-            real_writed(dos.tables.dpb,i*dos.tables.dpb_size+0x19,RealMake(dos.tables.dpb,(i+1)*dos.tables.dpb_size));
+            real_writed(dos.tables.dpb,i*dos.tables.dpb_size+(dos.version.major>=4?0x19:0x18),RealMake(dos.tables.dpb,(i+1)*dos.tables.dpb_size));
         else
-            real_writed(dos.tables.dpb,i*dos.tables.dpb_size+0x19,0xFFFFFFFF); // ED4.EXE (provided by Yksoft1) expects this, or else loops forever
+            real_writed(dos.tables.dpb,i*dos.tables.dpb_size+(dos.version.major>=4?0x19:0x18),0xFFFFFFFF); // ED4.EXE (provided by Yksoft1) expects this, or else loops forever
 	}
-    dos_infoblock.SetFirstDPB(RealMake(dos.tables.dpb,0));
+	dos_infoblock.SetFirstDPB(RealMake(dos.tables.dpb,0));
 
 	/* Create Device command packet area */
 	dos.dcp = DOS_GetMemory(3, "External device command packet");
@@ -431,6 +446,17 @@ void DOS_SetupTables(void) {
         dos.tables.country=country_info;
     }
 
+    /* Windows 95 FORMAT.COM does not use INT 21h to determine the boot drive. It uses
+     * the List of Lists and reads it directly from this infoblock. This value matches
+     * the hardcoded "C" drive the INT 21h emulation returns.
+     *
+     * No error checking of any kind is done by FORMAT.COM either, so if this value is
+     * zero, it will try to query for "@:\\WINBOOT.SYS" which of course does not exist.
+     *
+     * I'm given the impression from DOS_FindDevice() that someone probably ran into
+     * "@:" paths and did not understand this problem, which is fair enough. */
+    dos_infoblock.SetBootDrive(3); /* Drive C: (TODO: Make configurable, or else initially set to Z: then re-set to the first drive mounted */
+
     /* PC-98 INT 1Bh device list (60:6Ch-7Bh).
      * For now, just write a fake list to satisfy any PC-98 game that
      * requires a "master disk" to run even if running from an HDI.
