@@ -29,8 +29,10 @@
 #include "int10.h"
 #include "render.h"
 #include "dos_inc.h"
+#include "bitop.h"
 
 int hack_lfb_yadjust = 0;
+int hack_lfb_xadjust = 0;
 
 int vesa_set_display_vsync_wait = -1;
 bool vesa_bank_switch_window_range_check = true;
@@ -251,6 +253,7 @@ uint8_t VESA_GetSVGAModeInformation(uint16_t mode,uint16_t seg,uint16_t off) {
 	PhysPt buf=PhysMake(seg,off);
 	Bitu pageSize;
 	uint8_t modeAttributes;
+	unsigned int adj = 0;
 	Bitu i=0;
 
 	mode&=0x3fff;	// vbe2 compatible, ignore lfb and keep screen content bits
@@ -291,11 +294,14 @@ foundit:
 	bool allow_res = allow_vesa_lowres_modes ||
 		(ModeList_VGA[i].swidth >= 640 && ModeList_VGA[i].sheight >= 400);
 
+	unsigned int cwidth = (mblock->pitch != 0) ? mblock->pitch : mblock->swidth;
+
 	switch (mblock->type) {
 	case M_PACKED4:
-		if (!allow_vesa_4bpp_packed) return VESA_FAIL;//TODO: New option to disable
-		pageSize = mblock->sheight * mblock->swidth/2;
-		var_write(&minfo.BytesPerScanLine,(uint16_t)((((mblock->swidth+15U)/8U)&(~1U))*4)); /* NTS: 4bpp requires even value due to VGA registers, round up */
+		if (!allow_vesa_4bpp_packed && !(ModeList_VGA[i].mode >= 0x202 && ModeList_VGA[i].mode <= 0x208)) return VESA_FAIL;//TODO: New option to disable
+		pageSize = mblock->sheight * cwidth/2;
+		adj = hack_lfb_xadjust / 2;
+		var_write(&minfo.BytesPerScanLine,(uint16_t)((((cwidth+15U)/8U)&(~1U))*4)); /* NTS: 4bpp requires even value due to VGA registers, round up */
 		if (!int10.vesa_oldvbe10) { /* optional in VBE 1.0 */
 			var_write(&minfo.NumberOfPlanes,0x1);
 			var_write(&minfo.BitsPerPixel,4);
@@ -306,8 +312,9 @@ foundit:
 		break;
 	case M_LIN4:
 		if (!allow_vesa_4bpp) return VESA_FAIL;
-		pageSize = mblock->sheight * (uint16_t)(((mblock->swidth+15U)/8U)&(~1U));
-		var_write(&minfo.BytesPerScanLine,(uint16_t)(((mblock->swidth+15U)/8U)&(~1U))); /* NTS: 4bpp requires even value due to VGA registers, round up */
+		pageSize = mblock->sheight * (uint16_t)(((cwidth+15U)/8U)&(~1U));
+		adj = hack_lfb_xadjust / 8;
+		var_write(&minfo.BytesPerScanLine,(uint16_t)(((cwidth+15U)/8U)&(~1U))); /* NTS: 4bpp requires even value due to VGA registers, round up */
 		if (!int10.vesa_oldvbe10) { /* optional in VBE 1.0 */
 			var_write(&minfo.NumberOfPlanes,0x4);
 			var_write(&minfo.BitsPerPixel,4);   // bits per pixel is 4 as specified by VESA BIOS 2.0 specification
@@ -317,8 +324,9 @@ foundit:
 		break;
 	case M_LIN8:
 		if (!allow_vesa_8bpp || !allow_res) return VESA_FAIL;
-		pageSize = mblock->sheight * mblock->swidth;
-		var_write(&minfo.BytesPerScanLine,(uint16_t)mblock->swidth);
+		pageSize = mblock->sheight * cwidth;
+		adj = hack_lfb_xadjust;
+		var_write(&minfo.BytesPerScanLine,(uint16_t)cwidth);
 		if (!int10.vesa_oldvbe10) { /* optional in VBE 1.0 */
 			var_write(&minfo.NumberOfPlanes,0x1);
 			var_write(&minfo.BitsPerPixel,8);
@@ -329,8 +337,9 @@ foundit:
 		break;
 	case M_LIN15:
 		if (!allow_vesa_15bpp || !allow_res) return VESA_FAIL;
-		pageSize = mblock->sheight * mblock->swidth*2;
-		var_write(&minfo.BytesPerScanLine,(uint16_t)(mblock->swidth*2));
+		pageSize = mblock->sheight * cwidth*2;
+		adj = hack_lfb_xadjust * 2;
+		var_write(&minfo.BytesPerScanLine,(uint16_t)(cwidth*2));
 		if (!int10.vesa_oldvbe10) { /* optional in VBE 1.0 */
 			var_write(&minfo.NumberOfPlanes,0x1);
 			var_write(&minfo.BitsPerPixel,15);
@@ -349,8 +358,9 @@ foundit:
 		break;
 	case M_LIN16:
 		if (!allow_vesa_16bpp || !allow_res) return VESA_FAIL;
-		pageSize = mblock->sheight * mblock->swidth*2;
-		var_write(&minfo.BytesPerScanLine,(uint16_t)(mblock->swidth*2));
+		pageSize = mblock->sheight * cwidth*2;
+		adj = hack_lfb_xadjust * 2;
+		var_write(&minfo.BytesPerScanLine,(uint16_t)(cwidth*2));
 		if (!int10.vesa_oldvbe10) { /* optional in VBE 1.0 */
 			var_write(&minfo.NumberOfPlanes,0x1);
 			var_write(&minfo.BitsPerPixel,16);
@@ -368,8 +378,9 @@ foundit:
 	case M_LIN24:
 		if (!allow_vesa_24bpp || !allow_res) return VESA_FAIL;
 		if (mode >= 0x120 && !allow_explicit_vesa_24bpp) return VESA_FAIL;
-		pageSize = mblock->sheight * mblock->swidth*3;
-		var_write(&minfo.BytesPerScanLine,(uint16_t)(mblock->swidth*3));
+		pageSize = mblock->sheight * cwidth*3;
+		adj = hack_lfb_xadjust * 3;
+		var_write(&minfo.BytesPerScanLine,(uint16_t)(cwidth*3));
 		if (!int10.vesa_oldvbe10) { /* optional in VBE 1.0 */
 			var_write(&minfo.NumberOfPlanes,0x1);
 			var_write(&minfo.BitsPerPixel,24);
@@ -386,8 +397,9 @@ foundit:
 		break;
 	case M_LIN32:
 		if (!allow_vesa_32bpp || !allow_res) return VESA_FAIL;
-		pageSize = mblock->sheight * mblock->swidth*4;
-		var_write(&minfo.BytesPerScanLine,(uint16_t)(mblock->swidth*4));
+		pageSize = mblock->sheight * cwidth*4;
+		adj = hack_lfb_xadjust * 4;
+		var_write(&minfo.BytesPerScanLine,(uint16_t)(cwidth*4));
 		if (!int10.vesa_oldvbe10) { /* optional in VBE 1.0 */
 			var_write(&minfo.NumberOfPlanes,0x1);
 			var_write(&minfo.BitsPerPixel,32);
@@ -406,6 +418,7 @@ foundit:
 		break;
 	case M_TEXT:
 		if (!allow_vesa_tty) return VESA_FAIL;
+		adj = hack_lfb_xadjust / 8;
 		pageSize = 0;
 		var_write(&minfo.BytesPerScanLine, (uint16_t)(mblock->twidth * 2));
 		if (!int10.vesa_oldvbe10) { /* optional in VBE 1.0 */
@@ -482,7 +495,16 @@ foundit:
 		var_write(&minfo.YCharSize,(uint8_t)mblock->cheight); /* optional in VBE 1.0 */
 	}
 
-	if (!int10.vesa_nolfb && !int10.vesa_oldvbe) var_write(&minfo.PhysBasePtr,S3_LFB_BASE + (hack_lfb_yadjust*(long)host_readw((HostPt)(&minfo.BytesPerScanLine))));
+	/* this special hack is for the Windows 3.1 S3 driver in "640x480 (1MB) 16 million colors" mode which does no scan line or display offset calls at all */
+	if (mblock->special & _S3_POW2_STRIDE) {
+		uint16_t w = host_readw((HostPt)(&minfo.BytesPerScanLine));
+		if (!bitop::ispowerof2(w)) {
+			w = bitop::bit2mask(bitop::log2(w) + 1u);
+			var_write(&minfo.BytesPerScanLine,(uint16_t)w);
+		}
+	}
+
+	if (!int10.vesa_nolfb && !int10.vesa_oldvbe) var_write(&minfo.PhysBasePtr,S3_LFB_BASE + adj + (hack_lfb_yadjust*(long)host_readw((HostPt)(&minfo.BytesPerScanLine))));
 
 	MEM_BlockWrite(buf,&minfo,sizeof(MODE_INFO));
 	return VESA_SUCCESS;
@@ -584,6 +606,10 @@ uint8_t VESA_ScanLineLength(uint8_t subcall,uint16_t val, uint16_t & bytes,uint1
 	Bitu screen_height = CurMode->sheight;
 	Bitu max_offset;
 
+	/* this special hack is for the Windows 3.1 S3 driver in "640x480 (1MB) 16 million colors" mode which does no scan line or display offset calls at all */
+	if (CurMode->special & _S3_POW2_STRIDE)
+		return VESA_MODE_UNSUPPORTED;
+
 	switch (CurMode->type) {
 		case M_TEXT:
 			vmemsize = 0x8000;      // we have only the 32kB window here
@@ -696,6 +722,10 @@ uint8_t VESA_SetDisplayStart(uint16_t x,uint16_t y,bool wait) {
 	Bitu pixels_per_offset;
 	Bitu panning_factor = 1;
 
+	/* this special hack is for the Windows 3.1 S3 driver in "640x480 (1MB) 16 million colors" mode which does no scan line or display offset calls at all */
+	if (CurMode->special & _S3_POW2_STRIDE)
+		return VESA_MODE_UNSUPPORTED;
+
 	if (!wait) {
 		if (vesa_set_display_vsync_wait > 0)
 			wait = true;
@@ -851,6 +881,10 @@ Bitu INT10_WriteVESAModeList(Bitu max_modes) {
         else if (ModeList_VGA[i].type == M_ERROR) {
             /* ignore */
         }
+        /* do not list modes that are marked as do not list */
+        else if (ModeList_VGA[i].special & _DO_NOT_LIST) {
+            /* ignore */
+        }
         else {
             /* If there is no "accepts mode" then accept.
              *
@@ -882,6 +916,7 @@ Bitu INT10_WriteVESAModeList(Bitu max_modes) {
                         (ModeList_VGA[i].special & _USER_MODIFIED) ||
                         (ModeList_VGA[i].swidth <= SCALER_MAXWIDTH && ModeList_VGA[i].sheight <= SCALER_MAXHEIGHT);
                     bool allow_res = allow1 && allow2 && allow3 && allow4 && allow5;
+		    bool allow_s3_packed4 = (ModeList_VGA[i].mode >= 0x202 && ModeList_VGA[i].mode <= 0x208);
 
                     switch (ModeList_VGA[i].type) {
                         case M_LIN32:	canuse_mode=allow_vesa_32bpp && allow_res; break;
@@ -890,7 +925,7 @@ Bitu INT10_WriteVESAModeList(Bitu max_modes) {
                         case M_LIN15:	canuse_mode=allow_vesa_15bpp && allow_res; break;
                         case M_LIN8:	canuse_mode=allow_vesa_8bpp && allow_res; break;
                         case M_LIN4:	canuse_mode=allow_vesa_4bpp && allow_res; break;
-                        case M_PACKED4:	canuse_mode=allow_vesa_4bpp_packed && allow_res; break;
+                        case M_PACKED4:	canuse_mode=(allow_vesa_4bpp_packed || allow_s3_packed4) && allow_res; break;
                         case M_TEXT:	canuse_mode=allow_vesa_tty && allow_res; break;
                         default:	break;
                     }
