@@ -1,6 +1,6 @@
 /*
   Simple DirectMedia Layer
-  Copyright (C) 1997-2023 Sam Lantinga <slouken@libsdl.org>
+  Copyright (C) 1997-2024 Sam Lantinga <slouken@libsdl.org>
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -20,7 +20,7 @@
 */
 #include "../../SDL_internal.h"
 
-#if SDL_VIDEO_DRIVER_COCOA
+#ifdef SDL_VIDEO_DRIVER_COCOA
 
 #include "SDL_cocoavideo.h"
 
@@ -141,7 +141,6 @@ static long end_ticks = 0;        // inserted for DOSBox-X
         return;
     }
 
-    ime_incompos = 1; // inserted for DOSBox-X
     if (_markedText != aString) {
         _markedText = aString;
     }
@@ -154,8 +153,6 @@ static long end_ticks = 0;        // inserted for DOSBox-X
 
     DEBUG_IME(@"setMarkedText: %@, (%d, %d)", _markedText,
           selectedRange.location, selectedRange.length);
-    ime_incompos = 0;        //inserted for DOSBox-X
-    end_ticks = TickCount(); //inserted for DOSBox-X
 }
 
 - (void)unmarkText
@@ -178,7 +175,7 @@ static int GetEnableIME()
     CFBooleanRef ret = (CFBooleanRef)TISGetInputSourceProperty(is, kTISPropertyInputSourceIsASCIICapable);
     return CFBooleanGetValue(ret) ? 0 : 1;
 }
- 
+
 - (void)keyboardInputSourceChanged:(NSNotification *)notification
 {
     if(!GetEnableIME()) {
@@ -193,11 +190,8 @@ static int GetEnableIME()
     NSWindow *window = [self window];
     NSRect contentRect = [window contentRectForFrameRect:[window frame]];
     float windowHeight = contentRect.size.height;
-    //NSRect rect = NSMakeRect(_inputRect.x, windowHeight - _inputRect.y - _inputRect.h,
-    //                         _inputRect.w, _inputRect.h);
-    NSRect rect = NSMakeRect(_inputRect.x, windowHeight - _inputRect.y,
-                             _inputRect.w, 0); // Fixed for DOSBox-X
-
+    NSRect rect = NSMakeRect(_inputRect.x, windowHeight - _inputRect.y - _inputRect.h,
+                             _inputRect.w, _inputRect.h);
 
     if (actualRange) {
         *actualRange = aRange;
@@ -511,9 +505,19 @@ void Cocoa_HandleKeyEvent(_THIS, NSEvent *event)
     case NSEventTypeKeyUp:
         SDL_SendKeyboardKey(SDL_RELEASED, code);
         break;
-    case NSEventTypeFlagsChanged:
-        HandleModifiers(_this, code, (unsigned int)[event modifierFlags]);
+    case NSEventTypeFlagsChanged: {
+        // see if the new modifierFlags mean any existing keys should be pressed/released...
+        const unsigned int modflags = (unsigned int)[event modifierFlags];
+        HandleModifiers(_this, SDL_SCANCODE_LSHIFT, modflags);
+        HandleModifiers(_this, SDL_SCANCODE_LCTRL, modflags);
+        HandleModifiers(_this, SDL_SCANCODE_LALT, modflags);
+        HandleModifiers(_this, SDL_SCANCODE_LGUI, modflags);
+        HandleModifiers(_this, SDL_SCANCODE_RSHIFT, modflags);
+        HandleModifiers(_this, SDL_SCANCODE_RCTRL, modflags);
+        HandleModifiers(_this, SDL_SCANCODE_RALT, modflags);
+        HandleModifiers(_this, SDL_SCANCODE_RGUI, modflags);
         break;
+    }
     default: /* just to avoid compiler warnings */
         break;
     }
@@ -534,7 +538,7 @@ extern CGError CGSSetGlobalHotKeyOperatingMode(CGSConnection connection, CGSGlob
 
 void Cocoa_SetWindowKeyboardGrab(_THIS, SDL_Window * window, SDL_bool grabbed)
 {
-#if SDL_MAC_NO_SANDBOX
+#ifdef SDL_MAC_NO_SANDBOX
     CGSSetGlobalHotKeyOperatingMode(_CGSDefaultConnection(), grabbed ? CGSGlobalHotKeyDisable : CGSGlobalHotKeyEnable);
 #endif
 }
